class DeletableMixin:
    """Interface for classes that can be edited and deleted."""

    def delete(self):
        """Delete the object.

        Example usage:

        .. code-block:: python

            bot = credentialManager.bot("name")

            redditApp = bot.redditApp
            redditApp.delete()

            sentryToken = bot.sentryToken
            sentryToken.delete()

            databaseCredential = bot.databaseCredential
            databaseCredential.delete()

            userVerification = cre

        """
        self._credmgr.delete(f"{self._path}/{self.id}")
        del self



import json

from ..mixins import BaseModel, DeletableMixin, EditableMixin, OwnerMixin
from ..mixins.redditReddit import RedditAppMixin
from . import RedditApp
from .utils import resolveModelFromInput, resolveUser


class UserVerification(
    BaseModel, DeletableMixin, EditableMixin, OwnerMixin, RedditAppMixin
):
    _attrTypes = {
        **BaseModel._attrTypes,
        "id": "int",
        "user_id": "str",
        "redditor": "str",
        "reddit_app_id": "int",
        "extra_data": "dict",
        "owner_id": "int",
    }

    _editableAttrs = ["userId", "redditor", "redditAppId", "extraData"]
    _path = "/user_verifications"
    _credmgrCallable = "userVerification"
    _nameAttr = "userId"
    _nameMapping = {"user_id": "userId"}
    _canFetchByName = True
    _getByNamePath = "get_redditor"
    _apiNameMapping = {"userId": "user_id"}

    def __init__(self, credmgr, **kwargs):
        """Initialize an User Verification instance.

        User Verifications for associating an unique ID with a Redditor

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: ID of this User Verification.
        :param userId: An unique ID to associate with a Redditor.
        :param redditor: Redditor username to associate with an unique ID.
        :param redditAppId: ID of the RedditApp the Refresh Token is for.
        :param extraData: Extra data to be associated with this User Verification
        :param ownerId: ID of the `.User` that owns this User Verification

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(_credmgr, userId, redditApp, redditor=None, extraData=None, owner=None):
        """Create a new User Verification

        User Verifications for associating an unique ID with a Redditor

        :param str userId: An unique ID to associate with a Redditor (required)
        :param Union[RedditApp,int,str] redditApp: Reddit app the User Verification is
            for (required)
        :param str redditor: Redditor the User Verification is for. This is not usually
            set manually.
        :param dict extraData: Extra JSON data to include with verification
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to
            create for other users.

        :returns: UserVerification

        """

        data = {
            "user_id": userId,
            "reddit_app_id": resolveModelFromInput(_credmgr, RedditApp, redditApp),
        }
        if redditor:
            data["redditor"] = redditor
        if extraData:
            data["extra_data"] = json.dumps(extraData)
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/user_verifications", data=data)



from credmgr.models.utils import camelToSnake


class EditableMixin:
    _editableAttrs = []

    def edit(self, **kwargs):
        """Edit the object

        :param kwargs: The params to update on the object.

        :returns: The edited object.

        """
        payload = []

        for attr in self._editableAttrs:
            if attr in kwargs:
                if attr in self._apiNameMapping:
                    path = f"/{self._apiNameMapping[attr]}"
                else:
                    path = f"/{camelToSnake(attr)}"
                payload.append({"op": "replace", "path": path, "value": kwargs[attr]})
        self.__dict__.update(
            self._credmgr.patch(f"{self._path}/{self.id}", data=payload).__dict__
        )
        return self



import base64
import logging
from typing import Union

import asyncpraw
import praw
from praw.util.token_manager import BaseTokenManager

from ..mixins import BaseApp
from .utils import resolveModelFromInput, resolveUser

log = logging.getLogger(__package__)


class RedditApp(BaseApp):
    """A class for Reddit Apps.

    To obtain an instance of this class execute:

    .. code-block:: python

        bot = credmgr.bot("BotName")
        redditApp = bot.redditApp

    """

    _attrTypes = {
        **BaseApp._attrTypes,
        "client_id": "str",
        "client_secret": "str",
        "app_description": "str",
        "user_agent": "str",
        "app_type": "str",
        "redirect_uri": "str",
        "state": "str",
    }

    _editableAttrs = BaseApp._editableAttrs + [
        "clientId",
        "clientSecret",
        "appDescription",
        "userAgent",
        "appType",
        "redirectUri",
    ]
    _path = "/reddit_apps"
    _credmgrCallable = "redditApp"
    _reddit = None

    def __init__(self, credmgr, **kwargs):
        """Initialize a Reddit App instance.

        Reddit Apps are used for interacting with reddit

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param int id: ID of this Reddit App.
        :param str name: Name of this Reddit App.
        :param str clientId: Client ID of this Reddit App.
        :param str clientSecret: Client secret of the Reddit App.
        :param str appDescription: Description of what this Reddit App is used for.
        :param str userAgent: The user agent used to identify this Reddit App to Reddit.
        :param str appType: Type of app this Reddit App is. One of: ``web``,
            ``installed``, or ``script``.
        :param str redirectUri: URL that redditors are redirected to after authorizing
            this Reddit App to access their account.
        :param str state: Used to identify this Reddit App during the OAuth2 flow.
        :param int ownerId: ID of the `.User` that owns this Reddit App.

        """
        super().__init__(credmgr, **kwargs)

        if "state" in kwargs:
            self._fetched = True

    @staticmethod
    @resolveUser()
    def _create(
        _credmgr,
        name,
        clientId,
        userAgent,
        appType,
        redirectUri,
        clientSecret,
        appDescription,
        enabled,
        owner=None,
    ):
        """Create a new Reddit App

        Reddit Apps are used for interacting with Reddit

        :param str name: (required)
        :param str clientId: Client ID of the Reddit App (required)
        :param str userAgent: User agent used for requests to Reddit's API (required)
        :param str appType: Type of the app. One of `web`, `installed`, or `script`
            (required)
        :param str redirectUri: Redirect URI for Oauth2 flow. Defaults to user set
            redirect uri (required)
        :param str clientSecret: Client secret of the Reddit App
        :param str appDescription: Description of the Reddit App
        :param bool enabled: Allows the app to be used (defaults to `True`)
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for
            other users.

        :returns: RedditApp

        """
        data = {
            "app_name": name,
            "client_id": clientId,
            "user_agent": userAgent,
            "app_type": appType,
            "redirect_uri": redirectUri,
        }
        if clientSecret:
            data["client_secret"] = clientSecret
        if appDescription:
            data["app_description"] = appDescription
        if enabled:
            data["enabled"] = enabled
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/reddit_apps", data=data)

    def reddit(
        self,
        redditor=None,
        *,
        use_async=False,
        use_cache=True,
        reddit_class: Union[praw.Reddit, asyncpraw.Reddit, None] = None,
    ) -> praw.Reddit:
        """Provides an optionally authenticated [Async] PRAW

        :param str redditor: The redditor that you want the Reddit instance authorized
            as.
        :param bool use_async: Whether to return an asyncpraw instance.
        :param bool use_cache: Whether to fetch a new instance regardless if it was
            cached already.
        :param reddit_class: The Reddit class to use. Useful if you have a local version
            of PRAW.

        :returns: Reddit instance.

        """
        if not (isinstance(self._reddit, praw.Reddit) == (not use_async)):
            self._reddit = None
        if not (use_cache and self._reddit):
            if not reddit_class:
                if use_async:
                    reddit_class = asyncpraw.Reddit
                else:
                    reddit_class = praw.Reddit
            extra_kwargs = {}
            if redditor:
                extra_kwargs["token_manager"] = RefreshTokenManager(
                    self._credmgr, redditor, self
                )
            self._reddit = reddit_class(
                client_id=self.clientId,
                client_secret=self.clientSecret,
                user_agent=self.userAgent,
                redirect_uri=self.redirectUri,
                **extra_kwargs,
            )
        return self._reddit

    def genAuthUrl(self, scopes=None, permanent=False, userVerification=None):
        """Generates a URL for users to verify or authenticate their Reddit account

        :param Union[list,str] scopes: List of scopes needed. Pass ``'all'`` for all
            scopes. The ``identity`` scope will always be included. (default:
            ``['identity']``)
        :param bool permanent: Determines if a refresh token will be provided. (default:
            ``False``)
        :param Union[UserVerification,int,str] userVerification: Link to a
                :class:`.UserVerification`
                object. Accepted values are:

                - An :class:`.UserVerification` object
                - An :class:`.UserVerification` ``id``
                - An ``userId`` of a :class:`.UserVerification` record.

            If a :class:`.UserVerification` record does not exist, one will be created.

        :returns: Auth URL

        """
        from credmgr.models import UserVerification

        if scopes is None or scopes == "identity":
            scopes = ["identity"]
        elif scopes == "all":
            scopes = ["*"]
        if "identity" not in scopes and scopes != ["*"]:
            scopes = [scopes, "identity"]
        if permanent:
            duration = "permanent"
        else:
            duration = "temporary"
        uVerification = resolveModelFromInput(
            self._credmgr, UserVerification, userVerification, "userId"
        )
        if not uVerification and userVerification:
            uVerification = self._credmgr.userVerification.create(
                userVerification, self.id
            )
        if uVerification:
            state = base64.urlsafe_b64encode(
                f"{self.state}:{uVerification}".encode()
            ).decode()
        else:
            state = self.state
        return self.reddit().auth.url(scopes, state, duration)


class RefreshTokenManager(BaseTokenManager):
    def __init__(self, credmgr, redditor, reddit_app):
        self._credmgr = credmgr
        self.redditor = redditor
        self.reddit_app = reddit_app
        self.refreshToken = self._credmgr.refreshToken(
            self.redditor, self.reddit_app.id
        )
        super().__init__()

    def post_refresh_callback(self, authorizer):
        log.debug(
            f"Storing new refresh token for app: {self.reddit_app.name}({self.reddit_app.id}) redditor: u/{self.redditor}"
        )
        self.refreshToken.edit(refreshToken=authorizer.refresh_token)
        log.debug(
            f"Successfully sorted new refresh token for app: {self.reddit_app.name}({self.reddit_app.id}) redditor: u/{self.redditor}"
        )

    def pre_refresh_callback(self, authorizer):
        log.debug(
            f"Fetching refresh token for app: {self.reddit_app.name}({self.reddit_app.id}) redditor: u/{self.redditor}"
        )
        refreshToken = self._credmgr.refreshToken(self.redditor, self.reddit_app.id)
        log.debug(
            f"Fetched refresh token for app: {self.reddit_app.name}({self.reddit_app.id}) redditor: u/{self.redditor}"
        )
        if refreshToken:
            authorizer.refresh_token = refreshToken.refreshToken



import configparser
import os

from . import models
from .auth import ApiTokenAuth
from .config import Config
from .exceptions import InitializationError
from .models.utils import CachedProperty
from .requestor import Requestor, urljoin
from .serializer import Serializer

User = models.User
Bot = models.Bot
RedditApp = models.RedditApp
RefreshToken = models.RefreshToken
UserVerification = models.UserVerification
SentryToken = models.SentryToken
DatabaseCredential = models.DatabaseCredential


class CredentialManager(object):
    """The CredentialManager class provides a way to interact with CredentialManager's API.

    The official way to initialize an instance of this class is to:

    .. code-block:: python

        from credmgr import CredentialManager

        credentialManager = CredentialManager(apiToken="LIqbGjAeep3Ws5DH3LOEQPmw8UZ6ek")

    .. note::

        It is recommended to use environment variables or a ``.credmgr.ini`` file. See
        :ref:`Auth documentation <auth>` for more details.

    """

    _default = None
    _endpoint = "/api/v1"

    def __init__(
        self, configName=None, sessionClass=None, sessionKwargs=None, **kwargs
    ):
        """Initialize a CredentialManager instance.

        :param str configName: The name of a section in your ``.credmgr.ini`` file that
            credmgr will load its configuration is loaded from. If ``configName`` is
            ``None``, then it will look for it in the environment variable
            ``credmgr_configName``. If it is not found there, the ``default`` section is
            used.
        :param Session sessionClass: A Session class that will be used to create a
            requestor. If not set, use ``requests.Session`` (default: None).
        :param dict sessionKwargs: Dictionary with additional keyword arguments used to
            initialize the session (default: None).

        Additional keyword arguments will be used to initialize the :class:`.Config`
        object. This can be used to specify configuration settings during instantiation
        of the :class:`.CredentialManager` instance. For more details please see
        :ref:`configuration`.

        Required settings are:

        - apiToken

        OR

        - username
        - password

        .. warning::

            Using an API Token instead of a username/password is strongly recommended!

        The ``sessionClass`` and ``sessionKwargs`` allow for customization of the
        session :class:`.CredentialManager` will use. This allows, e.g., easily adding
        behavior to the requestor or wrapping its |Session|_ in a caching layer. Example
        usage:

        .. |Session| replace:: ``Session``

        .. _session: https://2.python-requests.org/en/master/api/#requests.Session

        .. code-block:: python

            import json, betamax, requests


            class JSONDebugRequestor(Requestor):
                def request(self, *args, **kwargs):
                    response = super().request(*args, **kwargs)
                    print(json.dumps(response.json(), indent=4))
                    return response


            mySession = betamax.Betamax(requests.Session())
            credentialManager = CredentialManager(
                ..., sessionClass=JSONDebugRequestor, session=mySession
            )

        """
        if sessionKwargs is None:
            sessionKwargs = {}
        configSection = None
        try:
            configSection = configName or os.getenv("credmgr_configName") or "DEFAULT"
            self.config = Config(configSection, **kwargs)
        except configparser.NoSectionError as exc:
            if configSection is not None:
                exc.message += "\nYou provided the name of a .credmgr.ini section that doesn't exist."
            raise

        self._server = urljoin(
            getattr(self.config, "server"), getattr(self.config, "endpoint")
        )
        apiToken = getattr(self.config, "apiToken")
        username = getattr(self.config, "username")
        password = getattr(self.config, "password")

        initErrorMessage = "Required settings are missing. Either 'apiToken' OR 'username' and 'password' must be specified, These settings can be provided in a .credmgr.ini file, as a keyword argument during the initialization of the `CredentialManager` class, or as an environment variable."
        if all([apiToken, username, password]):
            raise InitializationError(initErrorMessage)
        if apiToken:
            self._auth = ApiTokenAuth(apiToken)
        elif username and password:
            self._auth = (username, password)
        else:
            raise InitializationError(
                "API Token or a username/password pair must be set."
            )

        self._requestor = Requestor(
            self._server, self._auth, sessionClass, **sessionKwargs
        )
        self.serializer = Serializer(self)
        self._currentUser = None
        self._userDefaults = None
        self.getUserDefault = lambda key, default: self.userDefaults.get(key, default)
        self.user = models.UserHelper(self)
        """An instance of :class:`.UserHelper`.

        Provides an interface for interacting with :class:`.User`. For example, to get a
        :class:`.User` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            user = credentialManager.user(1)
            print(user.id)

        To create a :class:`.User` do:

        .. code-block:: python

            user = credentialManager.user.create(**userKwargs)

        See :meth:`~.UserHelper.create` for the required params.

        """
        self.bot = models.BotHelper(self)
        """An instance of :class:`.BotHelper`.

        Provides an interface for interacting with :class:`.Bot`. For example, to get a
        :class:`.Bot` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            bot = credentialManager.bot(1)
            print(bot.id)

        To create a :class:`.Bot` do:

        .. code-block:: python

            bot = credentialManager.bot.create(**botKwargs)

        See :meth:`~.BotHelper.create` for the required params.

        """

        self.redditApp = models.RedditAppHelper(self)
        """An instance of :class:`.RedditAppHelper`.

        Provides an interface for interacting with :class:`.RedditApp`. For example, to
        get a :class:`.RedditApp` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            redditApp = credentialManager.redditApp(1)
            print(redditApp.id)

        To create a :class:`.RedditApp` do:

        .. code-block:: python

            redditApp = credentialManager.redditApp.create(**redditAppKwargs)

        See :meth:`~.RedditAppHelper.create` for the required params.

        """

        self.refreshToken = models.RefreshTokenHelper(self)
        """An instance of :class:`.RefreshTokenHelper`.

        Provides an interface for interacting with :class:`.RefreshToken`. For example
        to get a :class:`.RefreshToken` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            refreshToken = credentialManager.refreshToken(1)
            print(refreshToken.id)

        .. note::

            Refresh tokens cannot be manually created.

        """

        self.userVerification = models.UserVerificationHelper(self)
        """An instance of :class:`.UserVerificationHelper`.

        Provides an interface for interacting with :class:`.UserVerification`. For
        example to get a :class:`.UserVerification` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            userVerification = credentialManager.userVerification(1)
            print(userVerification.id)

        To create a :class:`.UserVerification` do:

        .. code-block:: python

            userVerification = credentialManager.userVerification.create(**userVerificationKwargs)

        See :meth:`~.UserVerificationHelper.create` for the required params.

        """

        self.sentryToken = models.SentryTokenHelper(self)
        """An instance of :class:`.SentryTokenHelper`.

        Provides an interface for interacting with :class:`.SentryToken`. For example to
        get a :class:`.SentryToken` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            sentryToken = credentialManager.sentryToken(1)
            print(sentryToken.id)

        To create a :class:`.SentryToken` do:

        .. code-block:: python

            sentryToken = credentialManager.sentryToken.create(**sentryTokenKwargs)

        See :meth:`~.SentryTokenHelper.create` for the required params.

        """

        self.databaseCredential = models.DatabaseCredentialHelper(self)
        """An instance of :class:`.DatabaseCredentialHelper`.

        Provides an interface for interacting with :class:`.DatabaseCredential`. For
        example to get a :class:`.DatabaseCredential` with :attr:`id` of ``1`` you can
        do:

        .. code-block:: python

            databaseCredential = credentialManager.databaseCredential(1)
            print(databaseCredential.id)

        To create a :class:`.DatabaseCredential` do:

        .. code-block:: python

            databaseCredential = credentialManager.databaseCredential.create(
                **databaseCredentialKwargs
            )

        See :meth:`~.DatabaseCredentialHelper.create` for the required params.

        """

    def users(self, batchSize=10, limit=None):
        """List Users

        :param int batchSize: Number of Users to return in each batch (default: ``20``)
        :param int limit: Maximum number of Users to return

        :returns: A :class:`.Paginator` to iterate through the Users

        """
        return User(self).listItems(batchSize=batchSize, limit=limit)

    def bots(self, batchSize=20, limit=None, owner=None):
        """List Bots

        :param int batchSize: Number of Bots to return in each batch (default: ``20``)
        :param int limit: Maximum number of Bots to return
        :param Union[int,str,User] owner: Return Bots that are owner by this user

        :returns: A :class:`.Paginator` to iterate through the Bots

        """
        return Bot(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def redditApps(self, batchSize=20, limit=None, owner=None):
        """List RedditApps

        :param int batchSize: Number of RedditApps to return in each batch (default:
            ``20``)
        :param int limit: Maximum number of RedditApps to return
        :param Union[int,str,User] owner: Return RedditApps that are owner by this user

        :returns: A :class:`.Paginator` to iterate through the Reddit Apps

        """
        return RedditApp(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def refreshTokens(self, batchSize=20, limit=None, owner=None):
        """List RefreshTokens

        :param int batchSize: Number of RefreshTokens to return in each batch (default:
            ``20``)
        :param int limit: Maximum number of RefreshTokens to return
        :param Union[int,str,User] owner: Return RefreshTokens that are owned by this
            user

        :returns: A :class:`.Paginator` to iterate through the Refresh Tokens

        .. note::

            This is *not* the intended way to fetch refresh tokens. See:
            :meth:`~.RedditApp.reddit` for obtaining an authorized reddit instance.

        """
        return RefreshToken(self).listItems(
            batchSize=batchSize, limit=limit, owner=owner
        )

    def userVerifications(self, batchSize=20, limit=None, owner=None):
        """List UserVerifications

        :param int batchSize: Number of UserVerifications to return in each batch
            (default: ``20``)
        :param int limit: Maximum number of UserVerifications to return
        :param Union[int,str,User] owner: Return UserVerifications that are owned by
            this user

        :returns: A :class:`.Paginator` to iterate through the UserVerifications.

        """
        return UserVerification(self).listItems(
            batchSize=batchSize, limit=limit, owner=owner
        )

    def sentryTokens(self, batchSize=20, limit=None, owner=None):
        """List SentryTokens

        :param int batchSize: Number of SentryTokens to return in each batch (default:
            ``20``)
        :param int limit: Maximum number of SentryTokens to return
        :param Union[int,str,User] owner: Return SentryTokens that are owned by this
            user

        :returns: A :class:`.Paginator` to iterate through the SentryTokens

        """
        return SentryToken(self).listItems(
            batchSize=batchSize, limit=limit, owner=owner
        )

    def databaseCredentials(self, batchSize=20, limit=None, owner=None):
        """List DatabaseCredentials

        :param int batchSize: Number of DatabaseCredentials to return in each batch
            (default: ``20``)
        :param int limit: Maximum number of DatabaseCredentials to return
        :param Union[int,str,User] owner: Return DatabaseCredentials that are owned by
            this user

        :returns: A :class:`.Paginator` to iterate through the DatabaseCredentials

        """
        return DatabaseCredential(self).listItems(
            batchSize=batchSize, limit=limit, owner=owner
        )

    @CachedProperty
    def currentUser(self) -> User:
        """Returns the currently authenticated :class:`.User`"""
        if not self._currentUser:
            self._currentUser = self.get("/users/me")
        return self._currentUser

    @CachedProperty
    def userDefaults(self):
        """Returns the currently authenticated :class:`.User`'s default settings"""
        if not self._userDefaults:
            self._userDefaults = self.currentUser.defaultSettings
        return self._userDefaults

    def get(self, path, params=None):
        return self.serializer.deserialize(
            self._requestor.request(path, "GET", params=params)
        )

    def post(self, path, data):
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        return self.serializer.deserialize(
            self._requestor.request(path, "POST", data=data, headers=headers)
        )

    def patch(self, path, data):
        return self.serializer.deserialize(
            self._requestor.request(path, "PATCH", json=data)
        )

    def delete(self, path):
        return self._requestor.request(path, "DELETE")



from credmgr.mixins import BaseModel, DeletableMixin, EditableMixin, OwnerMixin
from credmgr.mixins.redditReddit import RedditAppMixin


class RefreshToken(
    BaseModel, DeletableMixin, OwnerMixin, EditableMixin, RedditAppMixin
):
    _attrTypes = {
        **BaseModel._attrTypes,
        "reddit_app_id": "int",
        "redditor": "str",
        "refresh_token": "str",
        "owner_id": "int",
        "scopes": "str",
        "issued_at": "datetime",
        "revoked": "bool",
        "revoked_at": "datetime",
    }
    _editableAttrs = ["refreshToken"]
    _path = "/refresh_tokens"
    _credmgrCallable = "refreshToken"
    _nameMapping = {"redditor": "redditor"}
    _apiNameMapping = {"redditor": "redditor", "redditAppId": "reddit_app_id"}
    _getByNamePath = "by_redditor"
    _canFetchByName = True
    _nameAttr = "redditor"

    def __init__(self, credmgr, **kwargs):
        """Initialize a Refresh Token instance.

        Refresh Tokens are for authenticating with Reddit as a Reddit that has
        authorized a `.RedditApp` to access their Reddit account.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param int id: ID of this Refresh Token.
        :param str redditor: Redditor this Refresh Token is for.
        :param str redditAppId: ID of the `.RedditApp` this Refresh Token is for.
        :param str refreshToken: The Refresh Token to pass to a Reddit instance.
        :param list[str] scopes: The OAuth2 scopes this Refresh Token grants access to.
        :param int ownerId: ID of the `.User` that owns this Refresh Token.
        :param datetime.datetime issuedAt: Date and time this Refresh Token was issued.
        :param bool revoked: Indicates if this Refresh Token was revoked or superseded
            by another Refresh Token.
        :param datetime.datetime revokedAt: Date and time this Refresh Token was revoked
            or superseded.

        """
        super().__init__(credmgr, **kwargs)



"""Provides the code to load credmgr's configuration file `~/.credmgr.ini`."""
import configparser
import os
import sys
from pathlib import Path
from threading import Lock


class Config:
    _config = None
    _lock = Lock()

    @staticmethod
    def _appendFileName(dirname):
        return os.path.join(dirname, ".credmgr.ini")

    @classmethod
    def _loadConfig(cls):
        """Attempt to load settings from various .credmgr.ini files."""
        config = configparser.ConfigParser()
        rootDir = os.path.dirname(sys.modules[__name__].__file__)
        locations = [
            cls._appendFileName(rootDir),
            cls._appendFileName(Path.home()),
            ".credmgr.ini",
        ]
        config.read(locations)
        cls._config = config

    def __init__(self, configName, **kwargs):
        """Initialize a Config instance."""
        with Config._lock:
            if Config._config is None:
                self._loadConfig()

        self._settings = kwargs
        self.customSettings = dict(
            Config._config.items(configName),
            **{k.lower(): v for k, v in kwargs.items() if v},
        )
        self.server = self.apiToken = self.username = self.password = None

        self._initializeConfig()

    def _fetch(self, key, default=None):
        configFileValue = self.customSettings.pop(key.lower(), default)
        envOverride = os.getenv(f"credmgr_{key}")
        return envOverride or configFileValue

    def _initializeConfig(self):
        self.server = self._fetch("server")
        self.endpoint = self._fetch("endpoint")
        self.apiToken = self._fetch("apitoken")
        self.username = self._fetch("username")
        self.password = self._fetch("password")
        self.dateFormat = self._fetch("dateformat")



import datetime
import re

import credmgr.models

from .exceptions import SerializerException


class Serializer(object):
    """The serializer builds :class:`.BaseModel` objects."""

    primitiveTypes = (float, bool, bytes, str, int)
    nativeTypeMapping = {
        "int": int,
        "str": str,
        "bool": bool,
        "datetime": datetime.datetime,
        "object": object,
        "dict": dict,
        "list": list,
    }

    def __init__(self, credmgr):
        """Initialize an Objector instance.

        :param credmgr: An instance of :class:`.CredentialManager`.

        """
        self._credmgr = credmgr

    @staticmethod
    def _snakeToCamel(snakeCase):
        return "".join(
            snakeCase.split("_")[:1]
            + [i.capitalize() for i in snakeCase.split("_")[1:]]
        )

    @staticmethod
    def deserializeDatatime(string):
        """Deserializes string to datetime.

        The string should be in iso8601 datetime format.

        :param string: str.

        :returns: datetime.

        """
        try:
            from dateutil.parser import parse

            return parse(string)
        except ImportError:  # pragma: no cover
            return string
        except ValueError:  # pragma: no cover
            raise SerializerException(f"Failed to parse `{string}` as datetime object")

    @staticmethod
    def deserializeDate(string):  # pragma: no cover
        """Deserializes string to date.

        :param string: str.

        :returns: date.

        """
        try:
            from dateutil.parser import parse

            return parse(string).date()
        except ImportError:
            return string
        except ValueError:
            raise SerializerException(f"Failed to parse `{string}` as date object")

    @staticmethod
    def deserializeObject(value):
        """Return a original value.

        :returns: object.

        """
        return value  # pragma: no cover

    @staticmethod
    def deserializePrimitive(data, objectType):
        """Deserializes string to primitive type.

        :param data: str.
        :param objectType: class literal.

        :returns: int, long, float, str, bool.

        """
        try:
            return objectType(data)
        except UnicodeEncodeError:  # pragma: no cover
            return str(data)
        except TypeError:  # pragma: no cover
            return data

    def deserialize(self, response):
        """Deserializes response into an object.

        :param response: Response object to be deserialized.

        :returns: deserialized object.

        """
        data = response.json()
        if isinstance(data, list) and all(
            [i["resource_type"] == data[0]["resource_type"] for i in data]
        ):
            if data:
                resourceType = f'list[{data[0]["resource_type"]}]'
            else:  # pragma: no cover
                return data
        else:
            resourceType = data.get("resource_type", "dict")
        return self.__deserialize(data, resourceType)

    def __deserialize(self, data, objectType):
        """Deserializes dict, list, str into an object.

        :param data: dict, list or str.
        :param objectType: class literal, or string of class name.

        :returns: object.

        """
        if data is None:
            return None

        if type(objectType) == str:
            if objectType.startswith("list["):
                subClass = re.match(r"list\[(.*)\]", objectType).group(1)
                return [self.__deserialize(subData, subClass) for subData in data]

            if objectType.startswith("dict"):
                return {k: self.__deserialize(v, type(v)) for k, v in data.items()}

            # convert str to class
            if objectType in self.nativeTypeMapping:
                objectType = self.nativeTypeMapping[objectType]
            else:
                objectType = getattr(credmgr.models, objectType)

        if objectType in self.primitiveTypes:
            return self.deserializePrimitive(data, objectType)
        elif objectType == object:  # pragma: no cover
            return self.deserializeObject(data)
        elif objectType == datetime.date:  # pragma: no cover
            return self.deserializeDate(data)
        elif objectType == datetime.datetime:
            return self.deserializeDatatime(data)
        elif objectType == dict:
            return self.__deserialize(data, "dict")
        else:
            return self.__deserializeModel(data, objectType)

    def __hasattr__(self, object, name):  # pragma: no cover
        return name in object.__class__.__dict__

    def __deserializeModel(self, data, objectType):
        """Deserializes list or dict to model.

        :param data: dict, list.
        :param objectType: class literal.

        :returns: model object.

        """

        if not objectType._attrTypes:  # pragma: no cover
            return data

        kwargs = {}
        if objectType._attrTypes is not None:
            for attr, attrType in objectType._attrTypes.items():
                if data and attr in data and isinstance(data, (list, dict)):
                    value = data[attr]
                    if attr in objectType._nameMapping:
                        kwargs[objectType._nameMapping[attr]] = self.__deserialize(
                            value, attrType
                        )
                    else:
                        kwargs[self._snakeToCamel(attr)] = self.__deserialize(
                            value, attrType
                        )

        try:
            instance = objectType(self._credmgr, **kwargs)
        except TypeError:  # pragma: no cover
            for key, value in kwargs.items():
                setattr(objectType, key, value)
            instance = objectType
        if (
            isinstance(instance, dict)
            and objectType._attrTypes is not None
            and isinstance(data, dict)
        ):  # pragma: no cover
            for key, value in data.items():
                if key not in objectType._attrTypes:
                    instance[key] = value
        return instance



from datetime import datetime

from ..exceptions import InitializationError
from ..models.utils import resolveUser


class BaseModel(object):
    """Superclass for all models in credmgr."""

    _attrTypes = {"id": "int"}
    _path = None
    _nameAttr = "name"
    _nameMapping = None
    _canFetchByName = False
    _credmgrCallable = None
    _getByNamePath = "by_name"
    _apiNameMapping = None

    def __init__(self, credmgr, **kwargs):
        """Initialize a BaseModel instance.

        :param credmgr: An instance of :class:`.CredentialManager`.

        """
        self._credmgr = credmgr
        if kwargs:
            for attribute, value in kwargs.items():
                setattr(self, attribute, value)
        self._fetched = False

    def __getattr__(self, attribute):  # pragma: no cover
        """Return the value of ``attribute``."""
        if not attribute.startswith("_") and not self._fetched:
            self._fetch()
            return getattr(self, attribute)
        raise AttributeError(
            f"{self.__class__.__name__!r} object has no attribute {attribute!r}"
        )

    def _get(self, id):
        self.__dict__ = self._credmgr.get(f"{self._path}/{id}").__dict__

    def _getByName(self):
        data = {}
        for modelAttr, dataAttr in self._apiNameMapping.items():
            name = getattr(self, modelAttr, None)
            if not name:  # pragma: no cover
                raise InitializationError(
                    f"Missing required keyword arg, {modelAttr!r}, to fetch object by name"
                )
            data[dataAttr] = name
        self.__dict__ = self._credmgr.post(
            f"{self._path}/{self._getByNamePath}", data=data
        ).__dict__

    def _fetch(self, byName=False):
        if byName and self._canFetchByName:
            self._getByName()
        else:
            self._get(self.id)
        self._fetched = True

    @resolveUser()
    def listItems(self, batchSize=20, limit=None, owner=None):
        from credmgr.models.helpers import Paginator

        return Paginator(
            self._credmgr,
            self.__class__,
            batchSize=batchSize,
            limit=limit,
            itemsOwner=owner,
        )

    def toDict(self):
        result = {}

        for exportAttr in self._attrTypes.keys():
            if exportAttr in self._nameMapping:
                attr = self._nameMapping.get(exportAttr)
            else:
                attr = "".join(
                    exportAttr.split("_")[:1]
                    + [i.capitalize() for i in exportAttr.split("_")[1:]]
                )
            value = getattr(self, attr, None)
            if value:
                if isinstance(value, list):
                    result[exportAttr] = list(
                        map(lambda x: x.toDict() if hasattr(x, "toDict") else x, value)
                    )
                elif hasattr(value, "toDict"):
                    result[exportAttr] = value.toDict()
                elif isinstance(value, dict):
                    result[exportAttr] = dict(
                        map(
                            lambda item: (item[0], item[1].toDict())
                            if hasattr(item[1], "toDict")
                            else item,
                            value.items(),
                        )
                    )
                elif isinstance(value, datetime):
                    result[exportAttr] = value.astimezone().strftime(
                        self._credmgr.config.dateFormat
                    )
                else:
                    result[exportAttr] = value
        if issubclass(type(self), dict):  # pragma: no cover
            for key, value in self.items():
                result[key] = value
        return result

    def __repr__(self):  # pragma: no cover
        return f"<{self.__class__.__name__} id={self.id}, {self._nameAttr}={getattr(self, self._nameAttr)!r}>"

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, type(self)):
            return False

        return self.id == other.id and getattr(self, self._nameAttr) == getattr(
            other, other._nameAttr
        )



from ..mixins import BaseApp
from .utils import resolveUser


class DatabaseCredential(BaseApp):
    _attrTypes = {
        **BaseApp._attrTypes,
        "database_username": "str",
        "database_host": "str",
        "database": "str",
        "database_flavor": "str",
        "database_port": "int",
        "database_password": "str",
        "use_ssh": "bool",
        "ssh_host": "str",
        "ssh_port": "int",
        "ssh_username": "str",
        "ssh_password": "str",
        "use_ssh_key": "bool",
        "private_key": "str",
        "private_key_passphrase": "str",
    }

    _editableAttrs = BaseApp._editableAttrs + [
        "databaseUsername",
        "databaseHost",
        "database",
        "databaseFlavor",
        "databasePort",
        "databasePassword",
        "useSsh",
        "sshHost",
        "sshPort",
        "sshUsername",
        "sshPassword",
        "useSshKey",
        "privateKey",
        "privateKeyPassphrase",
    ]
    _path = "/database_credentials"
    _credmgrCallable = "databaseCredential"

    def __init__(self, credmgr, **kwargs):
        """Initialize a Database Credential instance.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: The ID of the Database Credential.
        :param str name: Name of the Database Credential.
        :param str databaseFlavor: Type of database.
        :param str database: Working database to use.
        :param str databaseHost: Database server address.
        :param int databasePort: Port the database server listens on.
        :param str databaseUsername: Username used to connect and authenciate the
            database.
        :param str databasePassword: Password used to connect and authenciate the
            database.
        :param bool useSsh: Determines if the database will be connected to through a
            tunnel.
        :param str sshHost: The address of the server that the SSH tunnel will connect
            to.
        :param str sshPort: The port the SSH tunnel will use.
        :param str sshUsername: Username for the SSH tunnel.
        :param str sshPassword: Password for the SSH tunnel.
        :param bool useSshKey: Determines if the SSH tunnel will use private key
            authenciation.
        :param str privateKey: SSH private key. Note: No validation will be performed.
        :param str privateKeyPassphrase: Passphrase for the SSH key.
        :param ownerId: ID of the `.User` that owns this Database Credential.

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(
        _credmgr,
        name,
        databaseFlavor="postgres",
        database="postgres",
        databaseHost="localhost",
        databasePort=5432,
        databaseUsername="postgres",
        databasePassword=None,
        useSsh=False,
        sshHost=None,
        sshPort=None,
        sshUsername=None,
        sshPassword=None,
        useSshKey=False,
        privateKey=None,
        privateKeyPassphrase=None,
        enabled=True,
        owner=None,
    ):
        """Create a new Database Credential

        Database Credentials are used for..ya know..databases

        :param str name: Name of the Database Credential (required)
        :param str databaseFlavor: Type of database, (default: ``postgres``)
        :param str database: Working database to use, (default: ``postgres``)
        :param str databaseHost: Database server address, (default: ``localhost``)
        :param int databasePort: Port the database server listens on, (default:
            ``5432``)
        :param str databaseUsername: Username to use to connect to the database
        :param str databasePassword: Password to use to connect to the database
        :param bool useSsh: Determines if the database will be connected to through a
            tunnel
        :param str sshHost: The address of the server that the SSH tunnel will connect
            to
        :param str sshPort: The port the SSH tunnel will use
        :param str sshUsername: Username for the SSH tunnel
        :param str sshPassword: Password for the SSH tunnel
        :param bool useSshKey: Allows the credentials to be used
        :param str privateKey: SSH private key. Note: No validation will be performed.
        :param str privateKeyPassphrase: Passphrase for the SSH key
        :param bool enabled: Allows the credentials to be used
        :param Union[User,int,str] owner: Owner of the Database Credential. Requires
            Admin to create for other users.

        :returns: DatabaseCredential

        """

        data = {}
        if databaseFlavor:
            data["database_flavor"] = databaseFlavor
        if database:
            data["database"] = database
        if databaseHost:
            data["database_host"] = databaseHost
        if databasePort:
            data["database_port"] = databasePort
        if databaseUsername:
            data["database_username"] = databaseUsername
        if databasePassword:
            data["database_password"] = databasePassword
        if useSsh:
            data["use_ssh"] = useSsh
        if sshHost:
            data["ssh_host"] = sshHost
        if sshPort:
            data["ssh_port"] = sshPort
        if sshUsername:
            data["ssh_username"] = sshUsername
        if sshPassword:
            data["ssh_password"] = sshPassword
        if useSshKey:
            data["use_ssh_key"] = useSshKey
        if privateKey:
            data["private_key"] = privateKey
        if privateKeyPassphrase:
            data["private_key_passphrase"] = privateKeyPassphrase
        if enabled:
            data["enabled"] = enabled
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/database_credentials", data={"app_name": name, **data})



class CredentialManagerException(Exception):
    """The base CredentialManager Exception that all other exception classes extend."""


class InitializationError(CredentialManagerException):
    """Invalid initialization paramaters were used."""


class SerializerException(CredentialManagerException):
    """Serializaton/deserialzation error occured."""


class APIException(CredentialManagerException):
    """Base exception class"""

    def __init__(self, response):
        self.url = response.url
        self.status = response.status_code
        self.reason = response.reason
        self.body = response.text
        self.headers = response.headers

    def __str__(self):  # pragma: no cover
        """Custom error messages for exception"""
        errorMessage = f"\n{self.status}: {self.url}\n{self.reason}"
        if self.headers:
            errorMessage += f"\nResponse headers: {self.headers}"
        if self.body:
            errorMessage += f"\nResponse body: {self.body}"

        return errorMessage


class Conflict(APIException):
    """Made a conflicting request."""


class Forbidden(APIException):
    """Made request to a resource without required permissions."""


class InvalidJSON(APIException):
    """Response returned invalid JSON."""


class InvalidRequest(APIException):
    """Made request with invalid parameters."""


class NotFound(APIException):
    """Could not find requested resource."""


class ServerError(APIException):
    """Error occured on server during request."""


class Unauthorized(APIException):
    """Made request to resource without authenticating."""



import logging

import requests
from requests import codes

from .const import __version__
from .exceptions import (
    Conflict,
    Forbidden,
    InvalidJSON,
    InvalidRequest,
    NotFound,
    ServerError,
    Unauthorized,
)

log = logging.getLogger(__package__)


def urljoin(base, path):
    if base.endswith("/"):  # pragma: no cover
        base = base[:-1]
    if not path.startswith("/"):  # pragma: no cover
        path = f"/{path}"
    return f"{base}{path}"


class Requestor(object):
    """Requestor provides an interface to HTTP requests."""

    _exceptionMapping = {
        400: InvalidRequest,
        401: Unauthorized,
        403: Forbidden,
        404: NotFound,
        409: Conflict,
        422: ServerError,
        500: ServerError,
        502: ServerError,
        503: ServerError,
        504: ServerError,
        520: ServerError,
        522: ServerError,
    }
    _successCodes = [200, 201, 202, 204]

    def __init__(self, credmgrUrl, auth, sessionClass, **sessionKwargs):
        """Create an instance of the Requestor class.

        :param str credmgrUrl: Url used to make API requests
        :param auth: An auth tuple or a class that subclasses requests.auth
        :param Session sessionClass: (Optional) A custom session class to handle
            requests, compatible with requests.Session(). (Default: None)

        """
        self._baseUrl = credmgrUrl
        self._session = (sessionClass or requests.Session)()
        self._session.auth = auth
        for key, value in sessionKwargs.items():  # pragma: no cover
            setattr(self._session, key, value)
        self._session.headers["User-Agent"] = f"credmgr/{__version__}"

    def __getattr__(self, attribute):  # pragma: no cover
        """Pass all undefined attributes to the _session attribute."""
        if attribute.startswith("__"):
            raise AttributeError
        return getattr(self._session, attribute)

    @staticmethod
    def _logRequest(data, method, params, url):
        log.info(f"Data: {data}")
        log.info(f"Request: {method} {url}")
        log.info(f"Query Parameters: {params}")

    def request(self, path, method, data=None, params=None, **kwargs):
        url = urljoin(self._baseUrl, path)
        self._logRequest(data, method, params, url)
        response = self._session.request(method, url, params, data=data, **kwargs)
        if response.status_code in self._exceptionMapping:
            raise self._exceptionMapping[response.status_code](response)
        elif response.status_code == codes["no_content"]:
            return
        assert (
            response.status_code in self._successCodes
        ), f"Unexpected status code: {response.status_code}"
        if response.headers.get("content-length") == "0":  # pragma: no cover
            return ""
        try:
            response.json()
        except ValueError:  # pragma: no cover
            raise InvalidJSON(response)
        log.info(
            f'Response: {response.status_code} ({response.headers.get("content-length")} bytes)'
        )
        return response



import json

from ..exceptions import InitializationError
from ..mixins import BaseModel, DeletableMixin, EditableMixin


class User(BaseModel, DeletableMixin, EditableMixin):
    _attrTypes = {
        **BaseModel._attrTypes,
        "id": "int",
        "username": "str",
        "is_active": "bool",
        "is_regular_user": "bool",
        "is_admin": "bool",
        "default_settings": "dict(str, str)",
        "reddit_username": "str",
        "created": "datetime",
        "updated": "datetime",
        "reddit_apps": "list[RedditApp]",
        "sentry_tokens": "list[SentryToken]",
        "database_credentials": "list[DatabaseCredential]",
    }
    _editableAttrs = [
        "username",
        "isActive",
        "isRegularUser",
        "isAdmin",
        "defaultSettings",
        "redditUsername",
    ]
    _path = "/users"
    _credmgrCallable = "user"
    _nameAttr = "username"
    _nameMapping = {"username": "username"}
    _canFetchByName = True
    _apiNameMapping = {_nameAttr: "username"}

    def __init__(self, credmgr, **kwargs):
        """Initialize an User instance

        Users are for logging into and accessing that user's credentials

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param int id: ID of the User.
        :param username: Username of the User.
        :param bool isActive: Indicates if the User can log in and access
            CredentialManager.
        :param bool isRegularUser: Indicates if this is a regular user.
        :param bool isAdmin: Indicates if this User can create other users and their
            credentials.
        :param defaultSettings: Default settings used when creating new items without
            those settings explicitly set.
        :param str redditUsername: This User's Reddit username. Used for
            :class:`.RedditApp`'s userAgent.
        :param datetime.datetime created: Date and time this User was created.
        :param datetime.datetime updated: Date and time this User was last updated.
        :param list[RedditApp] redditApps: A list of Reddit Apps this User owns.
        :param list[SentryToken] sentryTokens: A list of Sentry Tokens this User owns.
        :param list[DatabaseCredential] databaseCredentials: A list of Database
            Credentials this User owns.

        """
        super().__init__(credmgr, **kwargs)
        self._apps = {}
        if "redditApps" in kwargs:
            self._apps["redditApps"] = kwargs["redditApps"]
        if "sentryTokens" in kwargs:
            self._apps["sentryTokens"] = kwargs["sentryTokens"]
        if "databaseCredentials" in kwargs:
            self._apps["databaseCredentials"] = kwargs["databaseCredentials"]

    @staticmethod
    def _create(
        _credmgr,
        username,
        password,
        defaultSettings=None,
        redditUsername=None,
        isAdmin=False,
        isActive=True,
        isRegularUser=True,
        isInternal=False,
    ):
        """Create a new User

        **PERMISSIONS: Admin role is required.**

        :param str username: Username for new user (Example: ``spaz``) (required)
        :param str password: Password for new user (Example: ``supersecurepassword``)
            (required)
        :param dict defaultSettings: Default values to use for new apps (Example:
            ``{"databaseFlavor": "postgres", "databaseHost": "localhost"}``)
        :param str redditUsername: User's Reddit username (Example: ``Lil_SpazJoekp``)
        :param bool isAdmin: Is the user an admin? Allows the user to see all objects
            and create users (Default: ``False``)
        :param bool isActive: Is the user active? Allows the user to sign in (Default:
            ``True``)
        :param bool isRegularUser: (Internal use only)
        :param bool isInternal: (Internal use only)

        :returns: User

        """
        additionalParams = {}
        if defaultSettings:
            additionalParams["default_settings"] = json.dumps(defaultSettings)
        if isAdmin:  # pragma: no cover
            additionalParams["is_admin"] = isAdmin
        if isActive:
            additionalParams["is_active"] = isActive
        if isRegularUser:
            additionalParams["is_regular_user"] = isRegularUser
        if isInternal:  # pragma: no cover
            additionalParams["is_internal"] = isInternal
        if redditUsername:
            additionalParams["reddit_username"] = redditUsername
        return _credmgr.post(
            "/users",
            data={"username": username, "password": password, **additionalParams},
        )

    def apps(self, only=None):
        """Returns apps that are owned by this user

        :param str only: Pass one of ``redditApps``, ``sentryTokens``,
            ``databaseCredentials`` to only get that type of apps

        :returns: Returns the user's apps.

        """
        appTypes = ["redditApps", "sentryTokens", "databaseCredentials"]
        if not self._apps:
            response = self._credmgr.get(f"/users/{self.id}/apps")
            self._apps = response._apps
            for app in appTypes:
                setattr(self, app, self._apps[app])
        if only:
            if only in appTypes:
                return self._apps[only]
            else:
                raise InitializationError(
                    f"App type: {only} is not valid. Only 'redditApps', 'sentryTokens', and 'databaseCredentials' are valid."
                )
        return self._apps



import re

from credmgr.exceptions import NotFound

reCamelToSnake = re.compile(r"([a-z0-9](?=[A-Z])|[A-Z](?=[A-Z][a-z]))")


def camelToSnake(name: str) -> str:
    """Convert `name` from camelCase to snake_case."""
    return reCamelToSnake.sub(r"\1_", name).lower()


def resolveUser(userAttr="owner", returnAttr="id"):
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            from . import User

            value = None
            user = kwargs.get(userAttr, None)
            if user:
                if isinstance(user, User):
                    value = getattr(user, returnAttr)
                elif isinstance(user, int):
                    value = user
                elif isinstance(user, str):
                    foundUser = self._credmgr.user(user)
                    if foundUser:
                        value = getattr(foundUser, returnAttr)
                kwargs[userAttr] = value
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


def resolveModelFromInput(credmgr, model, inputValue, returnAttr="id"):
    value = None
    if isinstance(inputValue, model):
        value = getattr(inputValue, returnAttr)
    elif isinstance(inputValue, int):
        value = inputValue
    elif isinstance(inputValue, str):
        try:
            foundItem = getattr(credmgr, model._credmgrCallable)(inputValue)
        except NotFound:
            foundItem = None
        if foundItem:
            value = getattr(foundItem, returnAttr)
    return value


class CachedProperty:
    def __init__(self, func, doc=None):
        self.func = self.__wrapped__ = func

        if doc is None:
            doc = func.__doc__
        self.__doc__ = doc

    # This to make sphinx run properly
    def __call__(self, *args, **kwargs):  # pragma: no cover noqa: D102
        pass

    def __get__(self, item, objtype=None):
        if item is None:
            return self

        value = item.__dict__[self.func.__name__] = self.func(item)
        return value



from ..mixins import BaseModel, DeletableMixin, EditableMixin, OwnerMixin


class BaseApp(BaseModel, DeletableMixin, EditableMixin, OwnerMixin):
    _attrTypes = {**BaseModel._attrTypes, "app_name": "str", "owner_id": "int"}
    _nameAttr = "name"
    _nameMapping = {"app_name": "name"}
    _editableAttrs = [
        "name",
    ]
    _apiNameMapping = {"name": "app_name"}



import re
from os import path

from setuptools import find_packages, setup

project_name = "credmgr"

root = path.abspath(path.dirname(__file__))

with open(path.join(root, project_name, "const.py")) as f:
    __version__ = re.search(r'__version__ = "(.+)"', f.read()).group(1)

with open(path.join(root, "README.rst")) as f:
    long_description = f.read()

extras = {
    "docs": ["sphinx", "sphinx_rtd_theme"],
    "lint": [
        "black",
        "docstrfmt",
        "flake8",
        "flynt",
        "isort",
        "pydocstyle",
        "sphinx<3.0",
        "sphinx_rtd_theme",
    ],
    "test": [
        "betamax",
        "betamax_serializers",
        "coverage",
        "mock",
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "pytest-xdist",
    ],
}
extras["dev"] = extras["docs"] + extras["lint"] + extras["test"]

requires = ["asyncpraw", "praw", "python-dateutil", "requests", "requests_toolbelt"]

setup(
    name=project_name,
    author="Lil_SpazJoekp",
    author_email="spaz@jesassn.org",
    license="Proprietary",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "License :: Other/Proprietary License",
    ],
    python_requires=">=3.6",
    description="Credential Manager API Client",
    include_package_data=True,
    install_requires=requires,
    long_description=long_description,
    package_data={"": ["LICENSE.txt"], project_name: ["*.ini"]},
    packages=find_packages(exclude=["tests", "tests.*"]),
    test_suite="tests",
    tests_require=extras["test"],
    extras_require=extras,
    url="https://credmgr.jesassn.org",
    version=__version__,
)



from ..mixins import BaseApp
from .utils import resolveUser


class SentryToken(BaseApp):
    _attrTypes = {**BaseApp._attrTypes, "dsn": "str"}
    _editableAttrs = BaseApp._editableAttrs + ["dsn"]
    _path = "/sentry_tokens"
    _credmgrCallable = "sentryToken"

    def __init__(self, credmgr, **kwargs):
        """Initialize a Sentry Token instance.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: ID of the Sentry Token
        :param str name: Name of the Sentry Token.
        :param str dsn: Sentry DSN url for reporting errors.
        :param ownerId: ID of the `.User` that owns this Sentry Token.

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(_credmgr, name=None, dsn=None, owner=None):
        """Create a new Sentry Token

        Sentry Tokens are used for logging and error reporting in applications

        :param str name: Name of the Sentry Token (required)
        :param str dsn: DSN of the Sentry Token (required)
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to
            create for other users.

        :returns: SentryToken

        """
        data = {"app_name": name, "dsn": dsn}
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/sentry_tokens", data=data)



from ..mixins import BaseApp
from .utils import resolveUser


class Bot(BaseApp):
    """A class for Bot

    To obtain an instance of this class execute:

    .. code-block:: python

        bot = credmgr.bot("botName")

    Bots is the intended way to use CredentialManager. Bots are for grouping credentials
    that would be used in a single app.

    The following is the intended usage:

    .. code-block:: python

        from credmgr import CredentialManager

        credentialManager = CredentialManager(apiToken="apiToken")
        bot = credmgr.bot("botName")

        reddit = bot.reddit("Lil_SpazJoekp")
        dbCreds = bot.databaseCredentials
        sentryDSN = bot.sentryToken

    """

    _attrTypes = {
        **BaseApp._attrTypes,
        "reddit_app": "RedditApp",
        "sentry_token": "SentryToken",
        "database_credential": "DatabaseCredential",
    }
    _editableAttrs = BaseApp._editableAttrs + [
        "redditAppId",
        "sentryTokenId",
        "databaseCredentialId",
    ]
    _path = "/bots"
    _credmgrCallable = "bot"
    _canFetchByName = True

    def __init__(self, credmgr, **kwargs):
        """Initialize a Bot instance.

        Bots are used for grouping credentials into a single app

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: ID of this Bot.
        :param name: Name of this Bot.
        :param ownerId: ID of the `.User` that owns this Bot.
        :param redditApp: `.RedditApp` that will be used with this Bot.
        :param sentryToken: `.SentryToken` that will be used with this Bot.
        :param databaseCredential: `.DatabaseCredential` that will be used with this
            Bot.

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @resolveUser()
    def _create(_credmgr, name, redditApp, sentryToken, databaseCredential, owner=None):
        """Create a new Bot

        Bots are used for grouping credentials into a single app

        :param str name: Name of the Bot (required)
        :param Union[RedditApp,int] redditApp: Reddit App the bot will use
        :param Union[SentryToken,int] sentryToken: Sentry Token the bot will use
        :param Union[DatabaseCredential,int] databaseCredential: Database Credentials
            the bot will use
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for
            other users.

        :returns: Bot

        """

        from . import DatabaseCredential, RedditApp, SentryToken

        additionalParams = {}
        if isinstance(redditApp, RedditApp):
            redditApp = redditApp.id
        if redditApp:
            additionalParams["reddit_app_id"] = redditApp
        if isinstance(sentryToken, SentryToken):
            sentryToken = sentryToken.id
        if sentryToken:
            additionalParams["sentry_token_id"] = sentryToken
        if isinstance(databaseCredential, DatabaseCredential):
            databaseCredential = databaseCredential.id
        if databaseCredential:
            additionalParams["database_credential_id"] = databaseCredential
        if owner:
            additionalParams["owner_id"] = owner
        return _credmgr.post("/bots", data={"app_name": name, **additionalParams})

    def edit(self, **kwargs):
        """:param name: Changes the name of the :class:`.Bot`
        :param Union[RedditApp,int] redditApp: Changes the :class:`.RedditApp` the Bot
            will use
        :param Union[SentryToken,int] sentryToken: Changes the :class:`.SentryToken` the
            Bot will use
        :param Union[DatabaseCredential,int] databaseCredential: Changes the
            :class:`.DatabaseCredential` the Bot will use

        .. note::

            Parameters, ``redditApp``, ``sentryToken``, and ``databaseCredential`` can
            accept the initialized object or its :attr:`id`. Passing a ``str`` to it
            will not work. The intended was to create bots in with the web interface.

        :returns: The modified :class:`.Bot`

        """
        from credmgr.models import DatabaseCredential, RedditApp, SentryToken

        iterKwargs = dict(kwargs.items())
        for key, value in iterKwargs.items():
            if key in ["redditApp", "sentryToken", "databaseCredential"]:
                if isinstance(
                    kwargs[key], (RedditApp, SentryToken, DatabaseCredential)
                ):
                    newKey = f"{key}Id"
                    kwargs[newKey] = kwargs.pop(key).id
                else:
                    newKey = f"{key}Id"
                    kwargs[newKey] = kwargs.pop(key)
        super(Bot, self).edit(**kwargs)



import logging

from ..exceptions import InitializationError
from ..mixins import BaseModel
from . import (
    Bot,
    DatabaseCredential,
    RedditApp,
    RefreshToken,
    SentryToken,
    User,
    UserVerification,
)
from .utils import resolveUser

log = logging.getLogger(__package__)


class Paginator:
    @resolveUser()
    def __init__(self, credmgr, model, batchSize=20, limit=None, itemsOwner=None):
        """Initialize a Paginator instance.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param model: A CredentialManager model to list.
        :param int batchSize: The number of items to fetch at a time. If ``batchSize``
            is None, it will fetch them 100 at a time. (default: 20).
        :param int limit: The maximum number of items to get.
        :param Union[int,User,str] itemsOwner: Owner to filter the items by.

        """
        self._credmgr = credmgr
        self._model = model(self._credmgr)
        self.batchSize = batchSize
        self.limit = limit
        self.itemsOwner = itemsOwner
        self._itemsReturned = 0
        self._completed = False
        self._offset = 0
        self._currentItemIndex = None
        self._response = None

    def __iter__(self):
        return self

    def __next__(self):
        if self.limit is not None and self._itemsReturned >= self.limit:
            raise StopIteration()  # pragma: no cover

        if self._response is None or self._currentItemIndex >= len(self._response):
            self._next()

        self._currentItemIndex += 1
        self._itemsReturned += 1
        return self._response[self._currentItemIndex - 1]

    def _next(self):
        if self._completed:
            raise StopIteration()
        params = dict(limit=self.batchSize, offset=self._offset)
        if self.itemsOwner:
            params["owner_id"] = self.itemsOwner
        self._response = self._credmgr.get(self._model._path, params=params)
        self._currentItemIndex = 0
        if not self._response:
            raise StopIteration()  # pragma: no cover
        if self._response and len(self._response) == self.batchSize:
            self._offset += self.batchSize  # pragma: no cover
        else:
            self._completed = True


class BaseHelper(BaseModel):
    _model = None

    def __call__(self, id=None, **kwargs):
        modelKwargs = {}
        byName = False
        if isinstance(id, str):
            name = id
            id = None
            if self._model._canFetchByName:
                byName = True
                kwargs[self._model._nameAttr] = name
            else:
                raise InitializationError(
                    f"Cannot get {self._model.__name__!r} by name"
                )
        if id:
            modelKwargs["id"] = id
        elif self._model._canFetchByName:
            byName = True
            for key in self._model._apiNameMapping.keys():
                name = kwargs.get(key, None)
                if name:
                    modelKwargs[key] = name
                else:
                    raise InitializationError(
                        f"Missing required keyword argument, {key!r}, to fetch object by name"
                    )
        else:
            raise InitializationError("'id' is required")
        item = self._model(self._credmgr, **modelKwargs)
        item._fetch(byName)
        return item


class UserHelper(BaseHelper):
    _model = User

    def __call__(self, id=None, username=None):
        """Fetches a :class:`.User` instance by :attr:`id`` or :attr:`name`.

        :param Union[int,str] id: ID of the :class:`.User` to fetch. .. note:: If a
            ``str`` is passed it will be treated as the :attr:`username` attr.
        :param str username: Username of the :class:`.User` to fetch. .. note:: If both
            :attr:`id` and :attr:`username` are passed the :attr:`id` will take
            precedence. If :attr:`id` is a ``str`` it will be treated as an
            :attr:`username` and will also take precedence.

        :returns: An initialized :class:`.User` instance.

        """
        kwargs = {}
        byName = False
        if isinstance(id, str):
            byName = True
            username = id
            id = None
        if id:
            kwargs["id"] = id
        if username:
            kwargs["username"] = username
        if not (id or username):
            raise InitializationError("At least 'id' or 'username' is required")
        item = User(self._credmgr, **kwargs)
        item._fetch(byName)
        return item

    def create(
        self,
        username,
        password,
        defaultSettings=None,
        isAdmin=False,
        isActive=True,
        isRegularUser=True,
        isInternal=False,
        redditUsername=None,
    ) -> User:
        """Create a new user.

        **PERMISSIONS: At least ``isAdmin`` is required.**

        :param str username: Username for new user (Example: ``spaz``) (required)
        :param str password: Password for new user (Example: ``supersecurepassword``)
            (required)
        :param dict defaultSettings: Default values to use for new apps (Example:
            ``{"databaseFlavor": "postgres", "databaseHost": "localhost"}``)
        :param bool isAdmin: Is the user an admin? Allows the user to see all objects
            and create users (Default: ``False``)
        :param bool isActive: Is the user active? Allows the user to sign in (Default:
            ``True``)
        :param bool isRegularUser: (Internal use only)
        :param bool isInternal: (Internal use only)
        :param str redditUsername: The user's Reddit username.

        :returns: User

        """

        return self._model._create(
            self._credmgr,
            username=username,
            password=password,
            defaultSettings=defaultSettings,
            isAdmin=isAdmin,
            isActive=isActive,
            isRegularUser=isRegularUser,
            isInternal=isInternal,
            redditUsername=redditUsername,
        )


class BotHelper(BaseHelper):
    _model = Bot

    def create(
        self,
        name,
        redditApp=None,
        sentryToken=None,
        databaseCredential=None,
        owner=None,
    ) -> Bot:
        """Create a new Bot.

        Bots are used for grouping credentials into a single app.

        :param str name: Name of the Bot (required).
        :param Union[RedditApp,int] redditApp: Reddit App the bot will use.
        :param Union[SentryToken,int] sentryToken: Sentry Token the bot will use.
        :param Union[DatabaseCredential,int] databaseCredential: Database Credentials
            the bot will use.
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for
            other users.

        :returns: Bot

        """

        return self._model._create(
            self._credmgr,
            name=name,
            redditApp=redditApp,
            sentryToken=sentryToken,
            databaseCredential=databaseCredential,
            owner=owner,
        )


class RedditAppHelper(BaseHelper):
    _model = RedditApp

    def create(
        self,
        name,
        clientId,
        userAgent=None,
        appType="web",
        redirectUri=None,
        clientSecret=None,
        appDescription=None,
        enabled=True,
        owner=None,
    ) -> RedditApp:
        """Create a new RedditApp.

        Reddit Apps are used for interacting with reddit.

        :param str name: Name of the RedditApp (required).
        :param str clientId: Client ID of the Reddit App (required).
        :param str userAgent: User agent used for requests to Reddit's API (required,
            defaults to user set default, then to 'python:{name} by /u/{redditUsername}'
            if currentUser.redditUsername is set or 'python:{name}' if it is not set).
        :param str appType: Type of the app. One of `web`, `installed`, or `script`
            (required, default: 'web').
        :param str redirectUri: Redirect URI for Oauth2 flow. (required, defaults to
            user set default then to
            `https://credmgr.jesassn.org/oauth2/reddit_callback` if neither are set).
        :param str clientSecret: Client secret of the Reddit App.
        :param str appDescription: Description of the Reddit App.
        :param bool enabled: Allows the app to be used.
        :param Union[User,int,str] owner: Owner of the Reddit App. Requires Admin to
            create for other users.

        :returns: RedditApp

        """
        if not userAgent:
            redditUsername = self._credmgr.currentUser.redditUsername
            redditUsernameStr = ""
            if redditUsername:
                redditUsernameStr = f" by /u/{redditUsername}"
            userAgent = self._credmgr.getUserDefault(
                "user_agent", f"python:{name}{redditUsernameStr}"
            )
        redirectUri = self._credmgr.getUserDefault(
            "redirect_uri",
            redirectUri or "https://credmgr.jesassn.org/oauth2/reddit_callback",
        )
        return self._model._create(
            self._credmgr,
            name=name,
            clientId=clientId,
            userAgent=userAgent,
            appType=appType,
            redirectUri=redirectUri,
            clientSecret=clientSecret,
            appDescription=appDescription,
            enabled=enabled,
            owner=owner,
        )


class UserVerificationHelper(BaseHelper):
    _model = UserVerification

    def create(
        self, userId, redditApp, redditor=None, extraData=None, owner=None
    ) -> UserVerification:
        """Create a new User Verification.

        User Verifications for verifying a redditor with a User ID.

        :param str userId: User ID to associate Redditor with (required).
        :param Union[RedditApp,int,str] redditApp: Reddit app the User Verification is
            for (required).
        :param str redditor: Redditor the User Verification is for.
        :param dict extraData: Extra JSON data to include with verification.
        :param int owner: Owner of the verification. Requires Admin to create for other
            users.

        :returns: UserVerification

        """
        return self._model._create(
            self._credmgr,
            userId=userId,
            redditApp=redditApp,
            redditor=redditor,
            extraData=extraData,
            owner=owner,
        )


class SentryTokenHelper(BaseHelper):
    _model = SentryToken

    def create(self, name, dsn, owner=None) -> SentryToken:
        """Create a new Sentry Token

        Sentry Tokens are used for logging and error reporting in applications

        :param str name: Name of the Sentry Token (required).
        :param str dsn: DSN of the Sentry Token (required).
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to
            create for other users.

        :returns: SentryToken

        """
        return self._model._create(self._credmgr, name=name, dsn=dsn, owner=owner)


class DatabaseCredentialHelper(BaseHelper):
    _model = DatabaseCredential

    def create(
        self,
        name,
        databaseFlavor="postgres",
        database="postgres",
        databaseHost="localhost",
        databasePort=5432,
        databaseUsername="postgres",
        databasePassword=None,
        useSsh=False,
        sshHost=None,
        sshPort=None,
        sshUsername=None,
        sshPassword=None,
        useSshKey=False,
        privateKey=None,
        privateKeyPassphrase=None,
        enabled=True,
        owner=None,
    ) -> DatabaseCredential:
        """Create a new Database Credential

        Database Credentials are used for..ya know..databases

        :param str name: Name of the Database Credential (required)
        :param str databaseFlavor: Type of database, (default: ``postgres``)
        :param str database: Working database to use, (default: ``postgres``)
        :param str databaseHost: Database server address, (default: ``localhost``)
        :param int databasePort: Port the database server listens on, (default:
            ``5432``)
        :param str databaseUsername: Username to use to connect to the database
        :param str databasePassword: Password to use to connect to the database
        :param bool useSSH: Determines if the database will be connected to through a
            tunnel
        :param str sshHost: The address of the server that the SSH tunnel will connect
            to
        :param str sshPort: The port the SSH tunnel will use
        :param str sshUsername: Username for the SSH tunnel
        :param str sshPassword: Password for the SSH tunnel
        :param bool useSSHKey: Allows the credentials to be used
        :param str privateKey: SSH private key. Note: No validation will be performed.
        :param str privateKeyPassphrase: Passphrase for the SSH key
        :param bool enabled: Allows the credentials to be used
        :param Union[User,int,str] owner: Owner of the app. Requires Admin to create for
            other users.

        :returns: DatabaseCredential

        """
        return self._model._create(
            self._credmgr,
            name=name,
            databaseFlavor=databaseFlavor,
            database=database,
            databaseHost=databaseHost,
            databasePort=databasePort,
            databaseUsername=databaseUsername,
            databasePassword=databasePassword,
            useSsh=useSsh,
            sshHost=sshHost,
            sshPort=sshPort,
            sshUsername=sshUsername,
            sshPassword=sshPassword,
            useSshKey=useSshKey,
            privateKey=privateKey,
            privateKeyPassphrase=privateKeyPassphrase,
            enabled=enabled,
            owner=owner,
        )


class RefreshTokenHelper(BaseHelper):
    _model = RefreshToken

    def __call__(self, id=None, redditor=None, redditAppId=None):
        kwargs = {}
        if isinstance(id, str):
            if redditor:
                redditAppId = redditor
            redditor = id
            id = None
        if id:
            kwargs["id"] = id
        if redditor:
            kwargs["redditor"] = redditor
        if redditAppId:
            kwargs["redditAppId"] = redditAppId
        return super().__call__(**kwargs)



from .base import BaseModel  # NOQA
from .deletable import DeletableMixin  # NOQA
from .editable import EditableMixin  # NOQA
from .owner import OwnerMixin  # NOQA

from .baseApp import BaseApp  # NOQA isort: skip
